/**
 * Logback: the reliable, generic, fast and flexible logging framework.
 * 
 * Copyright (C) 1999-2007, QOS.ch
 * 
 * This library is free software, you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation.
 */
package ch.qos.logback.eclipse.views;

import java.net.Socket;
import java.net.URL;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.osgi.util.NLS;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;

import ch.qos.logback.classic.boolex.JaninoEventEvaluator;
import ch.qos.logback.classic.spi.LoggingEvent;
import ch.qos.logback.core.filter.EvaluatorFilter;
import ch.qos.logback.core.spi.FilterReply;
import ch.qos.logback.eclipse.LogbackPlugin;
import ch.qos.logback.eclipse.Messages;
import ch.qos.logback.eclipse.PluginLogger;
import ch.qos.logback.eclipse.actions.LogbackFilterAction;
import ch.qos.logback.eclipse.actions.LogbackPreferencesAction;
import ch.qos.logback.eclipse.actions.OpenStackTraceAction;
import ch.qos.logback.eclipse.model.EventFilter;
import ch.qos.logback.eclipse.model.LoggingEventManager;
import ch.qos.logback.eclipse.reciever.SimpleSocketServer;
import ch.qos.logback.eclipse.util.EditorUtil;

/**
 * @author S&eacute;bastien Pennec
 */
public class LogbackView extends ViewPart {

  private static final String TAG_DISPLAY = "display"; //$NON-NLS-1$
  private static final String TAG_FONTNAME = "fontName"; //$NON-NLS-1$
  private static final String TAG_FONTSIZE = "fontSize"; //$NON-NLS-1$

  public static final String ID = "ch.qos.logback.eclipse.views.LogbackView"; //$NON-NLS-1$

  private TableViewer viewer;
  private TableColumn textColumn;
  private Integer tmpFontSize = null;
  private String tmpFontName = null;

  private LoggingEventContentProvider provider;
  private LoggingEventLabelProvider labelProvider;

  private Action clearConsoleAction;
  private Action autoScrollAction;

  private SimpleSocketServer server;
  private Thread serverThread;
  private LogbackPreferencesAction preferencesAction;
  private LogbackFilterAction filterAction;
  private OpenStackTraceAction openStackTraceAction;

  private IMemento memento;

  public LogbackView() {
  }

  public void createPartControl(Composite parent) {
    viewer = new TableViewer(parent, SWT.H_SCROLL | SWT.V_SCROLL | SWT.SINGLE | SWT.FULL_SELECTION
        | SWT.VIRTUAL);

    final Table table = viewer.getTable();

    textColumn = new TableColumn(table, SWT.LEFT);
    textColumn.setText("Event");
    textColumn.setWidth(2000);

    viewer.setItemCount(0);

    provider = new LoggingEventContentProvider(viewer);
    LoggingEventManager.getManager().addLoggingEventManagerListener(provider);
    viewer.setContentProvider(provider);

    labelProvider = new LoggingEventLabelProvider();
    if (memento != null) {
      labelProvider.init(memento);
      EventFilter.init(memento);
    }
    viewer.setLabelProvider(labelProvider);

    viewer.addDoubleClickListener(new IDoubleClickListener() {
      public void doubleClick(DoubleClickEvent e) {
        IStructuredSelection sel = (IStructuredSelection) e.getSelection();
        LoggingEvent event = (LoggingEvent) sel.getFirstElement();
        EditorUtil.openInEditor(event);
      }
    });

    makeActions();
    contributeToActionBars();
    createContextMenu();
    updateFont(tmpFontName, tmpFontSize);
  }

  private void contributeToActionBars() {
    IActionBars bars = getViewSite().getActionBars();
    IToolBarManager manager = bars.getToolBarManager();
    manager.add(clearConsoleAction);
    manager.add(autoScrollAction);
  }

  private void makeActions() {
    preferencesAction = new LogbackPreferencesAction(viewer, this);

    IMenuManager menu = getViewSite().getActionBars().getMenuManager();
    menu.add(preferencesAction);

    filterAction = new LogbackFilterAction(viewer);
    menu.add(filterAction);

    clearConsoleAction = new Action() {
      @Override
      public void run() {
        super.run();
        clearAll();
      }
    };
    clearConsoleAction.setText(Messages.ClearConsoleActionName);
    clearConsoleAction.setToolTipText(Messages.ClearConsoleActionTooltip);
    clearConsoleAction.setImageDescriptor(PlatformUI.getWorkbench().getSharedImages()
        .getImageDescriptor(ISharedImages.IMG_TOOL_DELETE));

    autoScrollAction = new Action() {
      @Override
      public void run() {
        super.run();
        provider.toggleAutoScroll();
        updateText();
        updateButton();
      }

      private void updateButton() {
        // TODO change the icon when autoscroll is toggled
      }

      private void updateText() {
        if (provider.getAutoScroll()) {
            this.setText(Messages.ScrollLockOffActionName);
            this.setToolTipText(Messages.ScrollLockOffActionTooltip);
        } else {
            this.setText(Messages.ScrollLockOnActionName);
            this.setToolTipText(Messages.ScrollLockOnActionTooltip);
        }
      }
    };
    autoScrollAction.setText(Messages.ScrollLockOffActionName);
    autoScrollAction.setToolTipText(Messages.ScrollLockOffActionTooltip);
    URL url = LogbackPlugin.getDefault().getBundle().getEntry("/icons/lock_co.gif"); //$NON-NLS-1$
    ImageDescriptor imgDesc = ImageDescriptor.createFromURL(url);
    autoScrollAction.setImageDescriptor(imgDesc);

    openStackTraceAction = new OpenStackTraceAction(this, Messages.OpenStacktraceActionName);

  }
  
  public void clearAll() {
    //public only for testing purpose
    LoggingEventManager.getManager().clearEventList();
    viewer.getTable().setItemCount(0);
    viewer.refresh();
  }

  @Override
  public void setFocus() {
    viewer.getControl().setFocus();
  }

  @Override
  public void dispose() {
    super.dispose();
    provider.dispose();
    labelProvider.dispose();
    server.stop();
  }

  public int getServerPort() {
    return server.getPort();
  }

  public String getPattern() {
    return labelProvider.getPattern();
  }

  public void updatePattern(String pattern) {
    labelProvider.updatePattern(pattern);
    viewer.refresh();
  }

  public void updateServerPort(Integer port) {
    if (port != null) {
      if (SimpleSocketServer.isPortFree(port)) {
        server.stop();
        openAndCloseClientToServer();
        if (serverThread != null) {
          serverThread.interrupt();
          serverThread = null;
        }
        server.setPort(port);
        startServerThread(server);
      }
    }
  }
  
  private void openAndCloseClientToServer() {
    try {
      new Socket("localhost", server.getPort()); //$NON-NLS-1$
    } catch (Exception e) {
      //do nothing
    }
  }

  private void initServer() {
    server = new SimpleSocketServer();
    if (memento != null) {
      server.init(memento);
    }
    startServerThread(server);
  }

  private void startServerThread(SimpleSocketServer server) {
    if (SimpleSocketServer.isPortFree(server.getPort())) {
      serverThread = new Thread(server);
      serverThread.start();
      PluginLogger.logInfo(NLS.bind(Messages.LogMessageListeningOnPort, server.getPort()));
    } else {
        MessageDialog.openInformation(this.getSite().getShell(), 
                Messages.CouldNotStartServerErrorDialogTitle, 
                NLS.bind(Messages.CouldNotStartServerError, server.getPort()));
    }
  }

  public void saveState(IMemento memento) {
    super.saveState(memento);
    EventFilter.saveState(memento);
    labelProvider.saveState(memento);
    server.saveState(memento);
    LoggingEventManager.getManager().saveState(memento);
    IMemento mem = memento.createChild(TAG_DISPLAY);
    FontData[] fd = viewer.getTable().getFont().getFontData();
    mem.putString(TAG_FONTNAME, fd[0].getName());
    mem.putInteger(TAG_FONTSIZE, fd[0].getHeight());
  }

  public void init(IViewSite site, IMemento memento) throws PartInitException {
    super.init(site, memento);
    this.memento = memento;
    initServer();
    if (memento != null) {
      LoggingEventManager.getManager().init(memento);
      
      IMemento mem = memento.getChild(TAG_DISPLAY);
      if (mem == null) {
        return;
      }

      tmpFontSize = mem.getInteger(TAG_FONTSIZE);
      tmpFontName = mem.getString(TAG_FONTNAME);
    }
  }

  public Viewer getViewer() {
    // used in tests
    return viewer;
  }

  private void createContextMenu() {
    MenuManager menuMgr = new MenuManager("#PopupMenu"); //$NON-NLS-1$
    menuMgr.setRemoveAllWhenShown(true);
    menuMgr.addMenuListener(new IMenuListener() {
      public void menuAboutToShow(IMenuManager m) {
        LogbackView.this.fillContextMenu(m);
      }
    });
    Menu menu = menuMgr.createContextMenu(viewer.getControl());
    viewer.getControl().setMenu(menu);
    getSite().registerContextMenu(menuMgr, viewer);
  }

  private void fillContextMenu(IMenuManager menuMgr) {
    boolean isEmpty = viewer.getSelection().isEmpty();
    boolean enable = false;
    if (!isEmpty) {
      IStructuredSelection selection = (IStructuredSelection) viewer.getSelection();
      LoggingEvent event = (LoggingEvent) selection.iterator().next();
      if (event.getThrowableProxy() != null) {
        enable = true;
      }
    }

    openStackTraceAction.setEnabled(enable);
    menuMgr.add(openStackTraceAction);
    menuMgr.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));
  }

  public LoggingEvent getSelectedEvent() {
    IStructuredSelection selection = (IStructuredSelection) viewer.getSelection();
    if (selection.size() > 0) {
      return (LoggingEvent) selection.iterator().next();
    }
    return null;
  }

  public Font getFont() {
    return viewer.getTable().getFont();
  }

  public void updateFont(String fontName, Integer fontSize) {
    if (fontName == null || fontSize == null) {
      return;
    }

    Font font = null;
    try {
      font = new Font(Display.getDefault(), fontName, fontSize, SWT.NORMAL);
    } catch (SWTError error) {
        PluginLogger.logInfo(Messages.LogMessageListeningOnPort);
    }
    if (font != null) {
      viewer.getTable().setFont(font);
      viewer.refresh();
    }
  }

  public String getLastLine() {
    //used for testing only
    if (viewer.getTable().getItemCount() == 0) {
      return null;
    }
    TableItem item = viewer.getTable().getItem((viewer.getTable().getItemCount()-1));
    return item.getText();
  }
  
  public int getNumberOfLines() {
    //used for testing only
    return viewer.getTable().getItemCount();
  }

  public void addNoDebugFilter() {
    //used for testing only
    EvaluatorFilter filter = FilterContentProvider.getProvider().createNewFilter();
    JaninoEventEvaluator eval = (JaninoEventEvaluator) filter.getEvaluator();
    eval.setExpression(Messages.DebugFilterExpression);
    eval.start();
    filter.setName(Messages.DebugFilterName);
    filter.setOnMatch(FilterReply.NEUTRAL);
    filter.setOnMismatch(FilterReply.DENY);
    filter.start();
    EventFilter.add(filter);
  }

}